package frc.robot.subsystems;

import java.util.function.BooleanSupplier;
import java.util.function.DoubleSupplier;

import com.revrobotics.RelativeEncoder;
import com.revrobotics.spark.SparkMax;
import com.revrobotics.spark.SparkBase.PersistMode;
import com.revrobotics.spark.SparkBase.ResetMode;
import com.revrobotics.spark.SparkLowLevel.MotorType;

import edu.wpi.first.math.MathUtil;
import edu.wpi.first.math.controller.ElevatorFeedforward;
import edu.wpi.first.math.controller.PIDController;
import edu.wpi.first.wpilibj.DigitalInput;
import edu.wpi.first.wpilibj2.command.Command;
import edu.wpi.first.wpilibj2.command.SubsystemBase;
import frc.robot.constants.ElevatorConstants;
import frc.robot.constants.ElevatorConstants.ElevatorControlMode;
import frc.robot.constants.ElevatorConstants.ElevatorPositions;

public class Elevator extends SubsystemBase {
    protected SparkMax elevatorMotor1;
    protected SparkMax elevatorMotor2;

    protected RelativeEncoder encoder;

    private BooleanSupplier hasAlgae;

    private DigitalInput bottomLimitSwitch;

    private PIDController pidController;

    private ElevatorFeedforward feedForward;
    private ElevatorFeedforward algaeHeldFeedForward;

    private ElevatorControlMode mode;

    private ElevatorPositions currentTargetPosition;
    
    public Elevator(BooleanSupplier hasAlgae) {
        elevatorMotor1 = new SparkMax(
            ElevatorConstants.kElevatorMotor1ID, 
            MotorType.kBrushless
        );
        
        elevatorMotor2 = new SparkMax(
            ElevatorConstants.kElevatorMotor2ID,
            MotorType.kBrushless
        );

        elevatorMotor1.configure(
            ElevatorConstants.motorConfig, 
            ResetMode.kResetSafeParameters, 
            PersistMode.kPersistParameters
        );

        elevatorMotor2.configure(
            ElevatorConstants.motorConfig.follow(ElevatorConstants.kElevatorMotor1ID), 
            ResetMode.kResetSafeParameters, 
            PersistMode.kPersistParameters
        );

        encoder = elevatorMotor1.getEncoder();

        this.hasAlgae = hasAlgae;

        bottomLimitSwitch = new DigitalInput(
            ElevatorConstants.kBottomLimitSwitchID
        );

        pidController = new PIDController(
            ElevatorConstants.kDownControllerP, 
            ElevatorConstants.kDownControllerI,
            ElevatorConstants.kDownControllerD
        );
        pidController.setSetpoint(0);

        pidController.setTolerance(ElevatorConstants.kAllowedError);

        feedForward = new ElevatorFeedforward(
            ElevatorConstants.kFeedForwardS, 
            ElevatorConstants.kFeedForwardG, 
            ElevatorConstants.kFeedForwardV
        );

        algaeHeldFeedForward = new ElevatorFeedforward(
            ElevatorConstants.kFeedForwardAlgaeHeldS, 
            ElevatorConstants.kFeedForwardAlgaeHeldG, 
            ElevatorConstants.kFeedForwardAlgaeHeldV
        );

        mode = ElevatorControlMode.kPID;

        currentTargetPosition = ElevatorPositions.kCoralIntake;
    }

    @Override
    public void periodic() {
        if (!getBottomLimitSwitch()) {
            encoder.setPosition(0);
        }

        if(mode == ElevatorControlMode.kPID) {
            if (!pidController.atSetpoint()) {
                elevatorMotor1.setVoltage(
                    pidController.calculate(
                        encoder.getPosition(),
                        currentTargetPosition.getPosition()
                    ) + (hasAlgae.getAsBoolean() ? algaeHeldFeedForward.calculate(0) : feedForward.calculate(0))
                );
            } else {
                if (hasAlgae.getAsBoolean()) {
                    elevatorMotor1.setVoltage(
                        algaeHeldFeedForward.calculate(0)
                    );
                } else {
                    elevatorMotor1.setVoltage(
                        feedForward.calculate(0)
                    );
                }
            }
        }
    }

    /**
     * Returns whether or not the motion is safe relative to the encoder's current position
     * and the elevator brace position
     * 
     * @return Is the motion safe
     */
    public boolean isMotionSafe() {
        return isMotionSafe(getEncoderPosition());
    }

    /**
     * Returns whether or not the motion is safe relative to some target position and the elevator
     * brace position
     * 
     * @param motionTarget The target position to determine the safety of
     * @return Is the motion safe
     */
    public boolean isMotionSafe(double motionTarget) {
        return motionTarget > ElevatorConstants.kBracePosition;
    }
    
    /**
     * A manual translation command that uses feed forward calculation to maintain position
     * 
     * @param speed The speed at which the elevator translates
     * @return Sets motor voltage to translate the elevator and maintain position
     */
    public Command runManualElevator(DoubleSupplier speed) {
        return run(() -> {
            double desired = speed.getAsDouble();

            if(Math.abs(MathUtil.applyDeadband(desired, .05)) > 0) {
                elevatorMotor1.set(
                    speed.getAsDouble()
                );
            } else {
                if (hasAlgae.getAsBoolean()) {
                    elevatorMotor1.setVoltage(
                        algaeHeldFeedForward.calculate(0)
                    );
                } else {
                    elevatorMotor1.setVoltage(
                        feedForward.calculate(0)
                    );
                }
            }
        });
    }

    public Command setTargetPosition(ElevatorPositions position) {
        return runOnce(() -> {
            if (mode == ElevatorControlMode.kManualMaintain) {
                pidController.reset();
                mode = ElevatorControlMode.kPID;
            }

            currentTargetPosition = position;
        });
    }

    public boolean atSetpoint() {
        return pidController.atSetpoint();
    }

    /**
     * Returns the current encoder position
     * 
     * @return Current encoder position
     */
    public double getEncoderPosition() {
        return encoder.getPosition();
    }


    /**
     * Returns the value of the bottom limit switch on the elevator (false = disabled, true = enabled)
     * 
     * @return The value of bottomLimitSwitch
     */
    public boolean getBottomLimitSwitch() {
        return bottomLimitSwitch.get();
    }

    /**
     * Returns the motor's output current
     * 
     * @return Motor output current
     */
    public double getMotor1() {
        return elevatorMotor1.getAppliedOutput()*elevatorMotor1.getBusVoltage();
    }

    /**
     * Returns the motor's output current
     * 
     * @return Motor output current
     */
    public double getMotor2() {
        return elevatorMotor2.getAppliedOutput()*elevatorMotor2.getBusVoltage();
    }

    public double getPIDSetpoint() {
        return pidController.getSetpoint();
    }

    public double getPIDError() {
        return pidController.getError();
    }
}